using System;
using System.Collections;
using System.IO;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Messaging;
using System.Runtime.Remoting.Channels;

using Belikov.GenuineChannels;
using Belikov.GenuineChannels.DirectExchange;
using Belikov.GenuineChannels.Messaging;
using Belikov.GenuineChannels.TransportContext;

namespace Server
{
	/// <summary>
	/// Represents a chat room.
	/// </summary>
	public class ChatRoom : IStreamResponseHandler
	{
		/// <summary>
		/// Contains a list of registered clients.
		/// Associates clients' URIs with HostInformation instances.
		/// </summary>
		private Hashtable _registeredClients = Hashtable.Synchronized(new Hashtable());

		/// <summary>
		/// Subscribes the client to chat room events.
		/// </summary>
		/// <param name="nickname">The client's nickname.</param>
		public void SubscribeClient(string nickname)
		{
			// subscribe it
			_registeredClients[GenuineUtility.CurrentRemoteUri] = GenuineUtility.CurrentRemoteHost;

			// and store its nickname
			GenuineUtility.CurrentRemoteHost["nickname"] = nickname;
		}

		/// <summary>
		/// Broadcasts the message to all subscribers.
		/// </summary>
		/// <param name="message">The message.</param>
		public void BroadcastMessage(string message)
		{
			string senderNickname = GenuineUtility.CurrentRemoteHost["nickname"] as string;
			if (senderNickname == null)
				senderNickname = "<unknown sender>";

			// write the message
			GenuineChunkedStream content = new GenuineChunkedStream(false);
			BinaryWriter binaryWriter = new BinaryWriter(content);
			binaryWriter.Write((string) message);
			binaryWriter.Write((string) senderNickname);

			ArrayList clientsBeingExcluded = new ArrayList();

			// send it to all clients
			lock (_registeredClients.SyncRoot)
			{
				foreach (DictionaryEntry dictionaryEntry in _registeredClients)
				{
					HostInformation hostInformation = null;

					try
					{
						hostInformation = (HostInformation) dictionaryEntry.Value;
						ChatServer.TransportContext.DirectExchangeManager.SendAsync(hostInformation, "ReceiveMessage", (Stream) content.Clone(), this, message);
					}
					catch(Exception ex)
					{
						Console.WriteLine("Remote receiver ({0}) is being excluded due to the error: {1}", hostInformation["nickname"], ex.Message);
						clientsBeingExcluded.Add(dictionaryEntry.Key);
					}
				}
			}

			foreach (object key in clientsBeingExcluded)
				_registeredClients.Remove(key);
		}

		/// <summary>
		/// Excludes clients from the client collection.
		/// </summary>
		/// <param name="response">The response.</param>
		/// <param name="remoteHost">The recipient of the message.</param>
		/// <param name="tag">The object provided during initiation of the invocation.</param>
		public void HandleException(Exception exception, HostInformation remoteHost, object tag)
		{
			string nickname = remoteHost["nickname"] as string;
			if (nickname == null)
				nickname = "<unknown>";
			Console.WriteLine("The client \"{0}\" did not receive the message \"{1}\" within the specified time span.", nickname, tag);

			this._registeredClients.Remove(remoteHost.Uri);
			Console.WriteLine("The client \"{0}\" was unsubscribed from chat events.", nickname);
		}

		/// <summary>
		/// Ignores received responses.
		/// </summary>
		/// <param name="response">The response.</param>
		/// <param name="remoteHost">The recipient of the message.</param>
		/// <param name="tag">The object provided during initiation of the invocation.</param>
		public void HandleResponse(Stream response, HostInformation remoteHost, object tag)
		{
			string nickname = remoteHost["nickname"] as string;
			if (nickname == null)
				nickname = "<unknown>";
			Console.WriteLine("The client \"{0}\" has confirmed that it received the message \"{1}\".", nickname, tag);
		}

	}
}
